@@ -14,8 +14,28 @@ from functools import wraps
 from inspect import getfullargspec, formatargspec, getcallargs
 
 from .encoders import BaseEncoder, UTF8Encoder
-from .exceptions import Error, ErrorReply, TransactionError, NotConnectedError, ConnectionLostError, NoRunningScriptError, ScriptKilledError
-from .replies import BlockingPopReply, DictReply, ListReply, PubSubReply, SetReply, StatusReply, ZRangeReply, ConfigPairReply, InfoReply, ClientListReply
+from .exceptions import (
+        Error,
+        ErrorReply,
+        TransactionError,
+        NotConnectedError,
+        ConnectionLostError,
+        NoRunningScriptError,
+        ScriptKilledError,
+)
+from .replies import (
+        BlockingPopReply,
+        ClientListReply,
+        ConfigPairReply,
+        DictReply,
+        EvalScriptReply,
+        InfoReply,
+        ListReply,
+        PubSubReply,
+        SetReply,
+        StatusReply,
+        ZRangeReply,
+)
 
 
 from .cursors import Cursor, SetCursor, DictCursor, ZCursor
@@ -55,11 +75,11 @@ class ZScoreBoundary:
 ZScoreBoundary.MIN_VALUE = ZScoreBoundary('-inf')
 ZScoreBoundary.MAX_VALUE = ZScoreBoundary('+inf')
 
+
 class ZAggregate: # TODO: use the Python 3.4 enum type.
     """
     Aggregation method for zinterstore and zunionstore.
     """
-
     #: Sum aggregation.
     SUM = 'SUM'
 
@@ -183,8 +203,9 @@ class PostProcessors:
                 ConfigPairReply: cls.multibulk_as_configpair,
                 ListOf(bool): cls.multibulk_as_boolean_list,
                 _ScanPart: cls.multibulk_as_scanpart,
+                EvalScriptReply: cls.any_to_evalscript,
 
-                None: None,  # XXX: We shouldn't allow 'None' anymore. It is still used by _subscribe.
+                NoneType: None,
         }[return_type]
 
     @classmethod
@@ -197,21 +218,21 @@ class PostProcessors:
             @asyncio.coroutine
             def as_list(protocol, result):
                 result = yield from original_post_processor(protocol, result)
-                return (yield from result.get_as_list())
+                return (yield from result.aslist())
             return '_aslist', list, as_list
 
         elif return_type == SetReply:
             @asyncio.coroutine
             def as_set(protocol, result):
                 result = yield from original_post_processor(protocol, result)
-                return (yield from result.get_as_set())
+                return (yield from result.asset())
             return '_asset', set, as_set
 
         elif return_type == DictReply:
             @asyncio.coroutine
             def as_dict(protocol, result):
                 result = yield from original_post_processor(protocol, result)
-                return (yield from result.get_as_dict())
+                return (yield from result.asdict())
             return '_asdict', dict, as_dict
 
     # === Post processor handlers below. ===
@@ -225,7 +246,7 @@ class PostProcessors:
     def multibulk_as_boolean_list(protocol, result):
         # Turn the array of integers into booleans.
         assert isinstance(result, MultiBulkReply)
-        values = yield from ListReply(result).get_as_list()
+        values = yield from ListReply(result).aslist()
         return [ bool(v) for v in values ]
 
     @asyncio.coroutine
@@ -246,13 +267,13 @@ class PostProcessors:
     @asyncio.coroutine
     def multibulk_as_blocking_pop_reply(protocol, result):
         assert isinstance(result, MultiBulkReply)
-        list_name, value = yield from ListReply(result).get_as_list()
+        list_name, value = yield from ListReply(result).aslist()
         return BlockingPopReply(list_name, value)
 
     @asyncio.coroutine
     def multibulk_as_configpair(protocol, result):
         assert isinstance(result, MultiBulkReply)
-        parameter, value = yield from ListReply(result).get_as_list()
+        parameter, value = yield from ListReply(result).aslist()
         return ConfigPairReply(parameter, value)
 
     @asyncio.coroutine
@@ -265,12 +286,12 @@ class PostProcessors:
         """
         # Get outer multi bulk reply.
         assert isinstance(result, MultiBulkReply)
-        new_cursor_pos, items_bulk = yield from ListReply(result).get_as_list()
+        new_cursor_pos, items_bulk = yield from ListReply(result).aslist()
         assert isinstance(items_bulk, MultiBulkReply)
 
         # Read all items for scan chunk in memory. This is fine, because it's
         # transmitted in chunks of about 10.
-        items = yield from ListReply(items_bulk).get_as_list()
+        items = yield from ListReply(items_bulk).aslist()
         return _ScanPart(int(new_cursor_pos), items)
 
     @asyncio.coroutine
@@ -318,8 +339,15 @@ class PostProcessors:
         assert isinstance(result, bytes)
         return float(result)
 
+    @asyncio.coroutine
+    def any_to_evalscript(protocol, result):
+        # Result can be native, int, MultiBulkReply or even a nested structure
+        assert isinstance(result, (int, bytes, MultiBulkReply, NoneType))
+        return EvalScriptReply(protocol, result)
+
 
 class ListOf:
+    """ Annotation helper for protocol methods. """
     def __init__(self, type_):
         self.type = type_
 
@@ -349,6 +377,9 @@ class CommandCreator:
     This wrapper handles (optionally) post processing of the returned data and
     implements some logic where commands behave different in case of a
     transaction or pubsub.
+
+    Warning: We use the annotations of `method` extensively for type checking
+             and determining which post processor to choose.
     """
     def __init__(self, method):
         self.method = method
@@ -360,6 +391,7 @@ class CommandCreator:
 
     @property
     def return_type(self):
+        """ Return type as defined in the method's annotation. """
         return self.specs.annotations.get('return', None)
 
     @property
@@ -384,7 +416,8 @@ class CommandCreator:
         else:
             return type_
 
-    def create_input_typechecker(self):
+    def _create_input_typechecker(self):
+        """ Return function that does typechecking on input data. """
         params = self.params
 
         if params:
@@ -406,7 +439,8 @@ class CommandCreator:
 
         return typecheck_input
 
-    def create_return_typechecker(self, return_type):
+    def _create_return_typechecker(self, return_type):
+        """ Return function that does typechecking on output data. """
         if return_type and not isinstance(return_type, str): # Exclude 'Transaction'/'Subscription' which are 'str'
             def typecheck_return(protocol, result):
                 """
@@ -444,12 +478,13 @@ class CommandCreator:
                     ClientListReply: ":class:`InfoReply <asyncio_redis.replies.ClientListReply>`",
                     ListReply: ":class:`ListReply <asyncio_redis.replies.ListReply>`",
                     MultiBulkReply: ":class:`MultiBulkReply <asyncio_redis.replies.MultiBulkReply>`",
-                    NativeType: "Native Python type, as defined by :attr:`encoder.native_type <asyncio_redis.encoders.BaseEncoder.native_type>`",
+                    NativeType: "Native Python type, as defined by :attr:`~asyncio_redis.encoders.BaseEncoder.native_type`",
                     NoneType: "None",
                     SetReply: ":class:`SetReply <asyncio_redis.replies.SetReply>`",
                     StatusReply: ":class:`StatusReply <asyncio_redis.replies.StatusReply>`",
                     ZRangeReply: ":class:`ZRangeReply <asyncio_redis.replies.ZRangeReply>`",
                     ZScoreBoundary: ":class:`ZScoreBoundary <asyncio_redis.replies.ZScoreBoundary>`",
+                    EvalScriptReply: ":class:`EvalScriptReply <asyncio_redis.replies.EvalScriptReply>`",
                     Cursor: ":class:`Cursor <asyncio_redis.cursors.Cursor>`",
                     SetCursor: ":class:`SetCursor <asyncio_redis.cursors.SetCursor>`",
                     DictCursor: ":class:`DictCursor <asyncio_redis.cursors.DictCursor>`",
@@ -469,6 +504,7 @@ class CommandCreator:
                     # XXX: Because of circulare references, we cannot use the real types here.
                     'Transaction': ":class:`asyncio_redis.Transaction`",
                     'Subscription': ":class:`asyncio_redis.Subscription`",
+                    'Script': ":returns: :class:`asyncio_redis.Script`",
                 }[type_]
             except KeyError:
                 if isinstance(type_, ListOf):
@@ -503,8 +539,8 @@ class CommandCreator:
         """
         Return the wrapped method for use in the `RedisProtocol` class.
         """
-        typecheck_input = self.create_input_typechecker()
-        typecheck_return = self.create_return_typechecker(return_type)
+        typecheck_input = self._create_input_typechecker()
+        typecheck_return = self._create_return_typechecker(return_type)
         method = self.method
 
         # Wrap it into a check which allows this command to be run either
@@ -583,6 +619,8 @@ class QueryCommandCreator(CommandCreator):
 
         return result
 
+_SMALL_INTS = list(str(i).encode('ascii') for i in range(1000))
+
 
 # List of all command methods.
 _all_commands = []
@@ -636,9 +674,10 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         transport, protocol = yield from loop.create_connection(RedisProtocol, 'localhost', 6379)
 
     :param password: Redis database password
-    :type password: bytes
+    :type password: Native Python type as defined by the ``encoder`` parameter
     :param encoder: Encoder to use for encoding to or decoding from redis bytes to a native type.
-    :type encoder: :class:`asyncio_redis.encoders.BaseEncoder` instance.
+                    (Defaults to :class:`~asyncio_redis.encoders.UTF8Encoder`)
+    :type encoder: :class:`~asyncio_redis.encoders.BaseEncoder` instance.
     :param db: Redis database
     :type db: int
     :param enable_typechecking: When ``True``, check argument types for all
@@ -727,7 +766,10 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
 
     def _encode_int(self, value:int) -> bytes:
         """ Encodes an integer to bytes. (always ascii) """
-        return str(value).encode('ascii')
+        if value < 1000: # For small values, take pre-encoded string.
+            return _SMALL_INTS[value]
+        else:
+            return str(value).encode('ascii')
 
     def _encode_float(self, value:float) -> bytes:
         """ Encodes a float to bytes. (always ascii) """
@@ -806,17 +848,9 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         """
         Coroutine which reads input from the stream reader and processes it.
         """
-        def item_callback(item):
-            f = self._queue.popleft()
-
-            if isinstance(item, Exception):
-                f.set_exception(item)
-            else:
-                f.set_result(item)
-
         while True:
             try:
-                yield from self._handle_item(item_callback)
+                yield from self._handle_item(self._push_answer)
             except ConnectionLostError:
                 return
 
@@ -886,7 +920,7 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
 
     @asyncio.coroutine
     def _handle_pubsub_multibulk_reply(self, multibulk_reply):
-        result = yield from ListReply(multibulk_reply).get_as_list()
+        result = yield from ListReply(multibulk_reply).aslist()
         assert result[0] in ('message', 'subscribe', 'unsubscribe', 'psubscribe', 'punsubscribe')
 
         if result[0] == 'message':
@@ -915,11 +949,11 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         # of arguments first.
 
         # Serialize and write header (number of arguments.)
-        data.append((u'*%i\r\n' % len(args)).encode('ascii'))
+        data += [ b'*', self._encode_int(len(args)), b'\r\n' ]
 
         # Write arguments.
         for arg in args:
-            data += [ (u'$%i\r\n' % len(arg)).encode('ascii'), arg, b'\r\n' ]
+            data += [ b'$', self._encode_int(len(arg)), b'\r\n', arg, b'\r\n' ]
 
         # Flush the last part
         self.transport.write(b''.join(data))
@@ -1103,11 +1137,6 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         (Returns 1 if the key was successfully renamed.) """
         return self._query(b'renamenx', self.encode_from_native(key), self.encode_from_native(newkey))
 
-#    @_query_command
-#    def getbit(self, key:NativeType, offset):
-#        """ Returns the bit value at offset in the string value stored at key """
-#        raise NotImplementedError
-
     @_query_command
     def bitop_and(self, destkey:NativeType, srckeys:ListOf(NativeType)) -> int:
         """ Perform a bitwise AND operation between multiple keys. """
@@ -1397,14 +1426,14 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         ::
 
             result = yield protocol.zrange('myzset', start=10, stop=20)
-            my_dict = yield result.get_as_dict()
+            my_dict = yield result.asdict()
 
         or the following to retrieve it as a list of keys:
 
         ::
 
             result = yield protocol.zrange('myzset', start=10, stop=20)
-            my_dict = yield result.get_as_list()
+            my_dict = yield result.aslist()
         """
         return self._query(b'zrange', self.encode_from_native(key),
                     self._encode_int(start), self._encode_int(stop), b'withscores')
@@ -1602,7 +1631,7 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
             yield from subscription.psubscribe(['pattern*'])
 
             while True:
-                result = yield from subscription.get_next_published()
+                result = yield from subscription.next_published()
                 print(result)
 
         :returns: :class:`asyncio_redis.Subscription`
@@ -1622,25 +1651,25 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         return subscription
 
     @_query_command
-    def _subscribe(self, channels:ListOf(NativeType)):
+    def _subscribe(self, channels:ListOf(NativeType)) -> NoneType:
         """ Listen for messages published to the given channels """
         self._pubsub_channels |= set(channels)
         return self._pubsub_method('subscribe', channels)
 
     @_query_command
-    def _unsubscribe(self, channels:ListOf(NativeType)):
+    def _unsubscribe(self, channels:ListOf(NativeType)) -> NoneType:
         """ Stop listening for messages posted to the given channels """
         self._pubsub_channels -= set(channels)
         return self._pubsub_method('unsubscribe', channels)
 
     @_query_command
-    def _psubscribe(self, patterns:ListOf(NativeType)):
+    def _psubscribe(self, patterns:ListOf(NativeType)) -> NoneType:
         """ Listen for messages published to channels matching the given patterns """
         self._pubsub_patterns |= set(patterns)
         return self._pubsub_method('psubscribe', patterns)
 
     @_query_command
-    def _punsubscribe(self, patterns:ListOf(NativeType)):
+    def _punsubscribe(self, patterns:ListOf(NativeType)) -> NoneType:
         """ Stop listening for messages posted to channels matching the given patterns """
         self._pubsub_patterns -= set(channels)
         return self._pubsub_method('punsubscribe', patterns)
@@ -1736,10 +1765,10 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
         """ Delete all the keys of the currently selected DB. This command never fails. """
         return self._query(b'flushdb')
 
-    @_query_command
-    def object(self, subcommand, args):
-        """ Inspect the internals of Redis objects """
-        raise NotImplementedError
+#    @_query_command
+#    def object(self, subcommand, args):
+#        """ Inspect the internals of Redis objects """
+#        raise NotImplementedError
 
     @_query_command
     def type(self, key:NativeType) -> StatusReply:
@@ -1805,8 +1834,8 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
 
     # LUA scripting
 
-    @_query_command
-    def register_script(self, script:str): # -> Script:
+    @_command
+    def register_script(self, script:str) -> 'Script':
         """
         Register a LUA script.
 
@@ -1814,8 +1843,6 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
 
             script = yield from protocol.register_script(lua_code)
             result = yield from script.run(keys=[...], args=[...])
-
-        :returns: :class:`asyncio_redis.Script`
         """
         # The register_script APi was made compatible with the redis.py library:
         # https://github.com/andymccurdy/redis-py
@@ -1846,22 +1873,21 @@ class RedisProtocol(asyncio.Protocol, metaclass=_RedisProtocolMeta):
     @_query_command
     def evalsha(self, sha:str,
                         keys:(ListOf(NativeType), NoneType)=None,
-                        args:(ListOf(NativeType), NoneType)=None):
+                        args:(ListOf(NativeType), NoneType)=None) -> EvalScriptReply:
         """
         Evaluates a script cached on the server side by its SHA1 digest.
         Scripts are cached on the server side using the SCRIPT LOAD command.
 
         The return type/value depends on the script.
         """
+        if not keys: keys = []
+        if not args: args = []
+
         try:
             result = yield from self._query(b'evalsha', sha.encode('ascii'),
                         self._encode_int(len(keys)),
                         *map(self.encode_from_native, keys + args))
 
-            # In case that we receive bytes, decode to string.
-            if isinstance(result, bytes):
-                result = self.decode_to_native(result)
-
             return result
         except ErrorReply as e:
             raise ScriptKilledError
@@ -2082,7 +2108,10 @@ class Script:
 
         ::
 
-            yield from script.run(keys=[], args=[])
+            script_reply = yield from script.run(keys=[], args=[])
+
+            # If the LUA script returns something, retrieve the return value
+            result = yield from script_reply.return_value()
         """
         return self.get_evalsha_func()(self.sha, keys, args)
 
@@ -2161,11 +2190,12 @@ class Subscription:
     def punsubscribe(self, patterns):
         return self.protocol._punsubscribe(self, patterns)
 
-    def get_next_published(self):
+    @asyncio.coroutine
+    def next_published(self):
         """
         Coroutine which waits for next pubsub message to be received and
         returns it.
 
         :returns: instance of :class:`PubSubReply <asyncio_redis.replies.PubSubReply>`
         """
-        return self._messages_queue.get()
+        return (yield from self._messages_queue.get())
